echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR icl 000 **************************"
echo ""
echo ""
copy /Y main.c_000 main.c
copy /Y decimal.h_000 decimal.h
copy /Y ..\LIBRARY\icl000libbid.lib ..\LIBRARY\libbid.lib
icl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR icl 001 **************************"
echo ""
echo ""
copy /Y main.c_001 main.c
copy /Y decimal.h_001 decimal.h
copy /Y ..\LIBRARY\icl001libbid.lib ..\LIBRARY\libbid.lib
icl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR icl 010 **************************"
echo ""
echo ""
copy /Y main.c_010 main.c
copy /Y decimal.h_010 decimal.h
copy /Y ..\LIBRARY\icl010libbid.lib ..\LIBRARY\libbid.lib
icl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR icl 011 **************************"
echo ""
echo ""
copy /Y main.c_011 main.c
copy /Y decimal.h_011 decimal.h
copy /Y ..\LIBRARY\icl011libbid.lib ..\LIBRARY\libbid.lib
icl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR icl 100 **************************"
echo ""
echo ""
copy /Y main.c_100 main.c
copy /Y decimal.h_100 decimal.h
copy /Y ..\LIBRARY\icl100libbid.lib ..\LIBRARY\libbid.lib
icl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR icl 101 **************************"
echo ""
echo ""
copy /Y main.c_101 main.c
copy /Y decimal.h_101 decimal.h
copy /Y ..\LIBRARY\icl101libbid.lib ..\LIBRARY\libbid.lib
icl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR icl 110 **************************"
echo ""
echo ""
copy /Y main.c_110 main.c
copy /Y decimal.h_110 decimal.h
copy /Y ..\LIBRARY\icl110libbid.lib ..\LIBRARY\libbid.lib
icl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR icl 111 **************************"
echo ""
echo ""
copy /Y main.c_111 main.c
copy /Y decimal.h_111 decimal.h
copy /Y ..\LIBRARY\icl111libbid.lib ..\LIBRARY\libbid.lib
icl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR cl 000 **************************"
echo ""
echo ""
copy /Y main.c_000 main.c
copy /Y decimal.h_000 decimal.h
copy /Y ..\LIBRARY\cl000libbid.lib ..\LIBRARY\libbid.lib
cl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR cl 001 **************************"
echo ""
echo ""
copy /Y main.c_001 main.c
copy /Y decimal.h_001 decimal.h
copy /Y ..\LIBRARY\cl001libbid.lib ..\LIBRARY\libbid.lib
cl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR cl 010 **************************"
echo ""
echo ""
copy /Y main.c_010 main.c
copy /Y decimal.h_010 decimal.h
copy /Y ..\LIBRARY\cl010libbid.lib ..\LIBRARY\libbid.lib
cl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR cl 011 **************************"
echo ""
echo ""
copy /Y main.c_011 main.c
copy /Y decimal.h_011 decimal.h
copy /Y ..\LIBRARY\cl011libbid.lib ..\LIBRARY\libbid.lib
cl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR cl 100 **************************"
echo ""
echo ""
copy /Y main.c_100 main.c
copy /Y decimal.h_100 decimal.h
copy /Y ..\LIBRARY\cl100libbid.lib ..\LIBRARY\libbid.lib
cl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR cl 101 **************************"
echo ""
echo ""
copy /Y main.c_101 main.c
copy /Y decimal.h_101 decimal.h
copy /Y ..\LIBRARY\cl101libbid.lib ..\LIBRARY\libbid.lib
cl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR cl 110 **************************"
echo ""
echo ""
copy /Y main.c_110 main.c
copy /Y decimal.h_110 decimal.h
copy /Y ..\LIBRARY\cl110libbid.lib ..\LIBRARY\libbid.lib
cl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR cl 111 **************************"
echo ""
echo ""
copy /Y main.c_111 main.c
copy /Y decimal.h_111 decimal.h
copy /Y ..\LIBRARY\cl111libbid.lib ..\LIBRARY\libbid.lib
cl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
del main.exe main.c decimal.h
del ..\LIBRARY\libbid.lib




echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR icl 000b **************************"
echo ""
echo ""
copy /Y main.c_000 main.c
copy /Y decimal.h_000 decimal.h
copy /Y ..\LIBRARY\icl000blibbid.lib ..\LIBRARY\libbid.lib
icl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR icl 001b **************************"
echo ""
echo ""
copy /Y main.c_001 main.c
copy /Y decimal.h_001 decimal.h
copy /Y ..\LIBRARY\icl001blibbid.lib ..\LIBRARY\libbid.lib
icl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR icl 010b **************************"
echo ""
echo ""
copy /Y main.c_010 main.c
copy /Y decimal.h_010 decimal.h
copy /Y ..\LIBRARY\icl010blibbid.lib ..\LIBRARY\libbid.lib
icl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR icl 011b **************************"
echo ""
echo ""
copy /Y main.c_011 main.c
copy /Y decimal.h_011 decimal.h
copy /Y ..\LIBRARY\icl011blibbid.lib ..\LIBRARY\libbid.lib
icl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR icl 100b **************************"
echo ""
echo ""
copy /Y main.c_100 main.c
copy /Y decimal.h_100 decimal.h
copy /Y ..\LIBRARY\icl100blibbid.lib ..\LIBRARY\libbid.lib
icl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR icl 101b **************************"
echo ""
echo ""
copy /Y main.c_101 main.c
copy /Y decimal.h_101 decimal.h
copy /Y ..\LIBRARY\icl101blibbid.lib ..\LIBRARY\libbid.lib
icl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR icl 110b **************************"
echo ""
echo ""
copy /Y main.c_110 main.c
copy /Y decimal.h_110 decimal.h
copy /Y ..\LIBRARY\icl110blibbid.lib ..\LIBRARY\libbid.lib
icl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR icl 111b **************************"
echo ""
echo ""
copy /Y main.c_111 main.c
copy /Y decimal.h_111 decimal.h
copy /Y ..\LIBRARY\icl111blibbid.lib ..\LIBRARY\libbid.lib
icl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR cl 000b **************************"
echo ""
echo ""
copy /Y main.c_000 main.c
copy /Y decimal.h_000 decimal.h
copy /Y ..\LIBRARY\cl000blibbid.lib ..\LIBRARY\libbid.lib
cl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR cl 001b **************************"
echo ""
echo ""
copy /Y main.c_001 main.c
copy /Y decimal.h_001 decimal.h
copy /Y ..\LIBRARY\cl001blibbid.lib ..\LIBRARY\libbid.lib
cl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR cl 010b **************************"
echo ""
echo ""
copy /Y main.c_010 main.c
copy /Y decimal.h_010 decimal.h
copy /Y ..\LIBRARY\cl010blibbid.lib ..\LIBRARY\libbid.lib
cl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR cl 011b **************************"
echo ""
echo ""
copy /Y main.c_011 main.c
copy /Y decimal.h_011 decimal.h
copy /Y ..\LIBRARY\cl011blibbid.lib ..\LIBRARY\libbid.lib
cl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR cl 100b **************************"
echo ""
echo ""
copy /Y main.c_100 main.c
copy /Y decimal.h_100 decimal.h
copy /Y ..\LIBRARY\cl100blibbid.lib ..\LIBRARY\libbid.lib
cl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR cl 101b **************************"
echo ""
echo ""
copy /Y main.c_101 main.c
copy /Y decimal.h_101 decimal.h
copy /Y ..\LIBRARY\cl101blibbid.lib ..\LIBRARY\libbid.lib
cl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR cl 110b **************************"
echo ""
echo ""
copy /Y main.c_110 main.c
copy /Y decimal.h_110 decimal.h
copy /Y ..\LIBRARY\cl110blibbid.lib ..\LIBRARY\libbid.lib
cl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
echo ""
echo ""
echo "***************** RUNNING EXAMPLE FOR cl 111b **************************"
echo ""
echo ""
copy /Y main.c_111 main.c
copy /Y decimal.h_111 decimal.h
copy /Y ..\LIBRARY\cl111blibbid.lib ..\LIBRARY\libbid.lib
cl main.c /DWINDOWS ..\LIBRARY\libbid.lib %1
main.exe
del main.exe main.c decimal.h
del ..\LIBRARY\libbid.lib
